@@ -10,6 +10,7 @@ use std::io;
 use anyhow::Context;
 use console::Term;
 use dialoguer::{theme, Confirm, Input};
+use futures::StreamExt;
 use itertools::Itertools;
 use serde::{de, Deserialize, Serialize};
 use serde_json as json;
@@ -69,7 +70,11 @@ pub(crate) struct Cli {
 enum Command {
     #[structopt(about = "Authenticate against statehub service", display_order(0))]
     Login,
-    #[structopt(about = "Create new state", aliases = &["create-st", "cs"], display_order(20))]
+    #[structopt(
+        about = "Create new state",
+        aliases = &["create-st", "cs"],
+        display_order(20),
+    )]
     CreateState {
         #[structopt(help = "State name")]
         name: v0::StateName,
@@ -86,22 +91,38 @@ enum Command {
         clusters: Vec<v0::ClusterName>,
     },
 
-    #[structopt(about = "Delete existing state", aliases = &["delete-st", "ds"], display_order(20))]
+    #[structopt(
+        about = "Delete existing state",
+        aliases = &["delete-st", "ds"],
+        display_order(20),
+    )]
     DeleteState {
         #[structopt(help = "State name")]
         name: v0::StateName,
     },
 
-    #[structopt(about = "List available states", aliases = &["list-state", "list-st", "ls"], display_order(20))]
+    #[structopt(
+        about = "List available states",
+        aliases = &["list-state", "list-st", "ls"],
+        display_order(20),
+    )]
     ListStates,
 
-    #[structopt(about = "Show state details", aliases = &["show-s", "ss"], display_order(20))]
+    #[structopt(
+        about = "Show state details",
+        aliases = &["show-s", "ss"],
+        display_order(20),
+    )]
     ShowState {
         #[structopt(help = "State name")]
         name: v0::StateName,
     },
 
-    #[structopt(about = "Register new cluster", aliases = &["register-cl", "rc"], display_order(10))]
+    #[structopt(
+        about = "Register new cluster",
+        aliases = &["register-cl", "rc"],
+        display_order(10),
+    )]
     RegisterCluster {
         #[structopt(
             help = "Non-interactive cluster registration, automatic defaults apply",
@@ -172,15 +193,30 @@ enum Command {
         cleanup_grace: String,
     },
 
-    #[structopt(about = "Unregister existing cluster", aliases = &["unregister-cl", "urc", "uc"], display_order(11))]
+    #[structopt(
+        about = "Unregister existing cluster",
+        aliases = &["unregister-cl", "urc", "uc"],
+        display_order(11),
+    )]
     UnregisterCluster {
         #[structopt(help = "Skip confirmation", long, short)]
         force: bool,
         #[structopt(help = "Cluster name")]
         name: v0::ClusterName,
+        #[structopt(
+            help = "Kubeconfig context to use for unregistering cluster",
+            long,
+            short,
+            visible_alias = "ctx"
+        )]
+        context: Option<String>,
     },
 
-    #[structopt(about = "Update registered cluster", aliases = &["update-cl", "upc"], display_order(12))]
+    #[structopt(
+        about = "Update registered cluster",
+        aliases = &["update-cl", "upc"],
+        display_order(12),
+    )]
     UpdateCluster {
         #[structopt(help = "Cluster name")]
         name: Option<v0::ClusterName>,
@@ -194,25 +230,41 @@ enum Command {
         azure: AzureCredentials,
     },
 
-    #[structopt(about = "List registered clusters", aliases = &["list-cluster", "list-cl", "lc"], display_order(13))]
+    #[structopt(
+        about = "List registered clusters",
+        aliases = &["list-cluster", "list-cl", "lc"],
+        display_order(13),
+    )]
     ListClusters,
 
-    #[structopt(about = "Show registered cluster details", aliases = &["show-cl", "sc"], display_order(14))]
+    #[structopt(
+        about = "Show registered cluster details",
+        aliases = &["show-cl", "sc"],
+        display_order(14),
+    )]
     ShowCluster {
         #[structopt(help = "Cluster name")]
         name: Option<v0::ClusterName>,
     },
 
-    #[structopt(about = "Diagnose registered cluster for any irregularities", aliases = &["diagnose-cl", "dc"], display_order(15))]
+    #[structopt(
+        about = "Diagnose registered cluster for any irregularities",
+        aliases = &["diagnose-cl", "dc"],
+        display_order(15),
+    )]
     DiagnoseCluster {
-        #[structopt(help = "Cluster name")]
-        name: Option<v0::ClusterName>,
+        #[structopt(
+            help = "Kubeconfig context to use for installing Statehub Cluster CRD",
+            long,
+            short
+        )]
+        context: Option<String>,
     },
 
     #[structopt(
         about = "Manually make state available in a specified location",
         aliases = &["add-l", "al"],
-        display_order(30)
+        display_order(30),
     )]
     AddLocation {
         #[structopt(help = "State name")]
@@ -238,7 +290,7 @@ enum Command {
     #[structopt(
         about = "Manually make state unavailable in a specified location",
         aliases = &["remove-l", "rem-l", "rl"],
-        display_order(30)
+        display_order(30),
     )]
     RemoveLocation {
         #[structopt(help = "State name")]
@@ -274,7 +326,7 @@ enum Command {
         about = "Manually create new volume",
         aliases = &["create-v", "cv"],
         display_order(50),
-        setting(clap::AppSettings::Hidden)
+        setting(clap::AppSettings::Hidden),
     )]
     CreateVolume {
         #[structopt(help = "State name")]
@@ -291,7 +343,7 @@ enum Command {
         about = "Manually delete existing volume",
         aliases = &["delete-v", "dv"],
         display_order(50),
-        setting(clap::AppSettings::Hidden)
+        setting(clap::AppSettings::Hidden),
     )]
     DeleteVolume {
         #[structopt(help = "State name")]
@@ -306,7 +358,7 @@ enum Command {
         about = "Manually choose primary location for volume",
         aliases = &["set-v", "sv"],
         display_order(50),
-        setting(clap::AppSettings::Hidden)
+        setting(clap::AppSettings::Hidden),
     )]
     SetVolume {
         #[structopt(help = "State name")]
@@ -320,24 +372,13 @@ enum Command {
     #[structopt(
         about = "List volumes in the given state",
         aliases = &["list-volume", "list-v", "lv"],
-        display_order(50)
+        display_order(50),
     )]
     ListVolumes {
         #[structopt(help = "State name")]
         state: v0::StateName,
     },
 
-    #[structopt(
-        about = "Create new namespace",
-        aliases = &["cns", "c-ns", "create-ns"],
-        display_order(1000),
-        setting(clap::AppSettings::Hidden)
-    )]
-    CreateNamespace {
-        #[structopt(help = "Namespace name", default_value = "statehub-system")]
-        namespace: String,
-    },
-
     #[structopt(
         about = "Save cluster token",
         alias = "sct",
@@ -355,7 +396,7 @@ enum Command {
         about = "Setup statehub configmap",
         aliases = &["scm", "setup-cm"],
         display_order(1000),
-        setting(clap::AppSettings::Hidden)
+        setting(clap::AppSettings::Hidden),
     )]
     SetupConfigmap {
         #[structopt(help = "cluster name")]
@@ -380,7 +421,7 @@ enum Command {
         about = "Install Statehub Cluster CRD",
         aliases = &["icc", "crd"],
         display_order(1000),
-        setting(clap::AppSettings::Hidden)
+        setting(clap::AppSettings::Hidden),
     )]
     InstallClusterCrd {
         #[structopt(help = "cluster name")]
@@ -398,6 +439,23 @@ enum Command {
         namespace: String,
     },
 
+    #[structopt(
+        about = "Uninstall Statehub Cluster CRD",
+        aliases = &["ucc", "uncrd"],
+        display_order(1000),
+        setting(clap::AppSettings::Hidden),
+    )]
+    UninstallClusterCrd {
+        #[structopt(help = "cluster name")]
+        cluster: v0::ClusterName,
+        #[structopt(
+            help = "Kubeconfig context to use for uninstalling Statehub Cluster CRD",
+            long,
+            short
+        )]
+        context: Option<String>,
+    },
+
     #[structopt(
         about = "List K8s namespaces",
         alias = "list-ns",
@@ -422,7 +480,8 @@ enum Command {
 
     #[structopt(
         about = "List K8s node regions",
-        aliases = &["list-r", "lr"], display_order(1000),
+        aliases = &["list-r", "lr"],
+        display_order(1000),
         setting(clap::AppSettings::Hidden),
     )]
     ListRegions {
@@ -433,7 +492,11 @@ enum Command {
     #[structopt(about = "Save default configuration file", display_order(2000))]
     SaveConfig,
 
-    #[structopt(about = "Check for updates to statehub utility", display_order(3000))]
+    #[structopt(
+        about = "Check for updates to statehub utility",
+        alias = "up",
+        display_order(3000)
+    )]
     SelfUpdate,
 }
 
@@ -534,8 +597,16 @@ impl Cli {
                         .await
                 }
             }
-            Command::UnregisterCluster { force, name } => {
-                statehub.unregister_cluster(name, force).await
+            Command::UnregisterCluster {
+                force,
+                name,
+                context,
+            } => {
+                let names = vec![name];
+                let contexts = context.into_iter().collect();
+                let kubectls = k8s::Kubectl::with_names_and_contexts(names, contexts)?;
+
+                statehub.unregister_cluster(kubectls, force).await
             }
             Command::UpdateCluster {
                 name,
@@ -562,13 +633,11 @@ impl Cli {
                 })?;
                 statehub.show_cluster(name).await
             }
-            Command::DiagnoseCluster { name } => {
-                let name = name.or_else(k8s::get_current_cluster_name).ok_or_else(|| {
-                    anyhow::anyhow!(
-                        "No default Kubernetes context found, need to provide cluster name"
-                    )
-                })?;
-                statehub.diagnose_cluster(name).await
+            Command::DiagnoseCluster { context } => {
+                let names = vec![];
+                let contexts = context.into_iter().collect();
+                let kubectls = k8s::Kubectl::with_names_and_contexts(names, contexts)?;
+                statehub.diagnose_cluster(kubectls).await
             }
             Command::AddLocation {
                 state,
@@ -607,7 +676,6 @@ impl Cli {
                 primary,
             } => statehub.set_volume_primary(state, volume, primary).await,
             Command::ListVolumes { state } => statehub.list_volumes(state).await,
-            Command::CreateNamespace { namespace } => statehub.create_namespace(namespace).await,
             Command::SaveClusterToken { namespace, token } => {
                 statehub.save_cluster_token(namespace, token).await
             }
@@ -631,6 +699,12 @@ impl Cli {
                 let kubectls = k8s::Kubectl::with_names_and_contexts(names, contexts)?;
                 statehub.install_cluster_crd(kubectls, namespace).await
             }
+            Command::UninstallClusterCrd { cluster, context } => {
+                let names = vec![cluster];
+                let contexts = context.into_iter().collect();
+                let kubectls = k8s::Kubectl::with_names_and_contexts(names, contexts)?;
+                statehub.uninstall_cluster_crd(kubectls).await
+            }
             Command::ListNamespaces => statehub.list_namespaces().await,
             Command::ListNodes => statehub.list_nodes().await,
             Command::ListPods => statehub.list_pods().await,
@@ -833,25 +907,41 @@ impl StateHub {
         }
     }
 
-    async fn diagnose_cluster(&self, name: v0::ClusterName) -> anyhow::Result<()> {
-        self.inform(format_args!("Running diagnostics on K8s context {}", name))?;
+    async fn diagnose_cluster(&self, kubectls: Vec<k8s::Kubectl>) -> anyhow::Result<()> {
+        for kubectl in kubectls {
+            self.inform(format_args!(
+                "Running diagnostics on cluster '{}' (K8s context '{}')",
+                kubectl.cluster_name(),
+                kubectl.context(),
+            ))?;
+            let check_cluster = kubectl.check_cluster();
+            futures::pin_mut!(check_cluster);
+            while let Some(item) = check_cluster.next().await {
+                match item {
+                    Ok(text) => self.inform(text)?,
+                    Err(text) => {
+                        self.inform(text)?;
+                        break;
+                    }
+                }
+            }
+        }
         Ok(())
     }
 
-    async fn unregister_cluster(&self, name: v0::ClusterName, force: bool) -> anyhow::Result<()> {
+    async fn unregister_cluster(
+        &self,
+        kubectls: Vec<k8s::Kubectl>,
+        force: bool,
+    ) -> anyhow::Result<()> {
         if force || self.confirm("Are you sure?") {
             log::info!("Make sure all the pods using statehub are terminated");
             log::info!("Uninstall helm");
-
-            self.relinquish_states_helper(&name).await?;
-
-            self.api
-                .unregister_cluster(name)
-                .await
-                .print(&self.stdout, self.json)
-        } else {
-            Ok(())
+            for kubectl in kubectls {
+                self.unregister_cluster_helper(kubectl).await?;
+            }
         }
+        Ok(())
     }
 
     async fn add_location(
@@ -1003,12 +1093,6 @@ impl StateHub {
         output.into_text(self.json)
     }
 
-    async fn create_namespace(&self, namespace: String) -> anyhow::Result<()> {
-        k8s::validate_namespace(&namespace)
-            .await
-            .map(|namespace| println!("{:#?}", namespace))
-    }
-
     async fn save_cluster_token(&self, namespace: String, token: String) -> anyhow::Result<()> {
         let secret = k8s::store_cluster_token(&namespace, &token).await?;
         if let Some(token) = k8s::extract_cluster_token(&secret) {
@@ -1045,7 +1129,15 @@ impl StateHub {
         namespace: String,
     ) -> anyhow::Result<()> {
         for kubectl in kubectls {
-            kubectl.install_cluster_crd(&namespace).await?;
+            kubectl.install_statehub_cluster(&namespace).await?;
+        }
+
+        Ok(())
+    }
+
+    async fn uninstall_cluster_crd(&self, kubectls: Vec<k8s::Kubectl>) -> anyhow::Result<()> {
+        for kubectl in kubectls {
+            kubectl.uninstall_statehub_cluster().await?;
         }
 
         Ok(())
